"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
If you make changes to this file, run "python -m dagster_cloud.pex.grpc.compile" after."""
import builtins
import google.protobuf.descriptor
import google.protobuf.message
import sys

if sys.version_info >= (3, 8):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing_extensions.final
class Empty(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___Empty = Empty

@typing_extensions.final
class CreatePexServerRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    CREATE_PEX_SERVER_ARGS_FIELD_NUMBER: builtins.int
    create_pex_server_args: builtins.str
    def __init__(
        self,
        *,
        create_pex_server_args: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["create_pex_server_args", b"create_pex_server_args"]) -> None: ...

global___CreatePexServerRequest = CreatePexServerRequest

@typing_extensions.final
class CreatePexServerReply(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    CREATE_PEX_SERVER_RESPONSE_FIELD_NUMBER: builtins.int
    create_pex_server_response: builtins.str
    def __init__(
        self,
        *,
        create_pex_server_response: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["create_pex_server_response", b"create_pex_server_response"]) -> None: ...

global___CreatePexServerReply = CreatePexServerReply

@typing_extensions.final
class GetPexServersRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    GET_PEX_SERVERS_ARGS_FIELD_NUMBER: builtins.int
    get_pex_servers_args: builtins.str
    def __init__(
        self,
        *,
        get_pex_servers_args: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["get_pex_servers_args", b"get_pex_servers_args"]) -> None: ...

global___GetPexServersRequest = GetPexServersRequest

@typing_extensions.final
class GetPexServersReply(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    GET_PEX_SERVERS_RESPONSE_FIELD_NUMBER: builtins.int
    get_pex_servers_response: builtins.str
    def __init__(
        self,
        *,
        get_pex_servers_response: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["get_pex_servers_response", b"get_pex_servers_response"]) -> None: ...

global___GetPexServersReply = GetPexServersReply

@typing_extensions.final
class GetCrashedPexServersRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    GET_CRASHED_PEX_SERVERS_ARGS_FIELD_NUMBER: builtins.int
    get_crashed_pex_servers_args: builtins.str
    def __init__(
        self,
        *,
        get_crashed_pex_servers_args: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["get_crashed_pex_servers_args", b"get_crashed_pex_servers_args"]) -> None: ...

global___GetCrashedPexServersRequest = GetCrashedPexServersRequest

@typing_extensions.final
class GetCrashedPexServersReply(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    GET_CRASHED_PEX_SERVERS_RESPONSE_FIELD_NUMBER: builtins.int
    get_crashed_pex_servers_response: builtins.str
    def __init__(
        self,
        *,
        get_crashed_pex_servers_response: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["get_crashed_pex_servers_response", b"get_crashed_pex_servers_response"]) -> None: ...

global___GetCrashedPexServersReply = GetCrashedPexServersReply

@typing_extensions.final
class ShutdownPexServerRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SHUTDOWN_PEX_SERVER_ARGS_FIELD_NUMBER: builtins.int
    shutdown_pex_server_args: builtins.str
    def __init__(
        self,
        *,
        shutdown_pex_server_args: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["shutdown_pex_server_args", b"shutdown_pex_server_args"]) -> None: ...

global___ShutdownPexServerRequest = ShutdownPexServerRequest

@typing_extensions.final
class ShutdownPexServerReply(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SHUTDOWN_PEX_SERVER_RESPONSE_FIELD_NUMBER: builtins.int
    shutdown_pex_server_response: builtins.str
    def __init__(
        self,
        *,
        shutdown_pex_server_response: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["shutdown_pex_server_response", b"shutdown_pex_server_response"]) -> None: ...

global___ShutdownPexServerReply = ShutdownPexServerReply

@typing_extensions.final
class PingRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ECHO_FIELD_NUMBER: builtins.int
    echo: builtins.str
    def __init__(
        self,
        *,
        echo: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["echo", b"echo"]) -> None: ...

global___PingRequest = PingRequest

@typing_extensions.final
class PingReply(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ECHO_FIELD_NUMBER: builtins.int
    echo: builtins.str
    def __init__(
        self,
        *,
        echo: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["echo", b"echo"]) -> None: ...

global___PingReply = PingReply
