from dagster_cloud_cli.core.headers.auth import (
    API_TOKEN_HEADER as API_TOKEN_HEADER,
    AUTHORIZATION_HEADER as AUTHORIZATION_HEADER,
    BEARER_HEADER_SCHEME as BEARER_HEADER_SCHEME,
    DEPLOYMENT_NAME_HEADER as DEPLOYMENT_NAME_HEADER,
)
